#ifndef HARKTOOL4_INCREMENTER_HPP
#define HARKTOOL4_INCREMENTER_HPP

#include <limits>
#include "utils_std.hpp"

namespace harktool4
{

/** @brief 一定範囲を、指定した増加量によって増加する増分器.
 * 範囲 [min_value, max_value) を、incrementずつ増加していく。
 * 初期値はbeginであり、インクリメントによってmax_value以上になると、
 * (max_value - min_value)が減算され、範囲内に収まる。
 * 現在の値は*演算子で取得できる。
 * incrementに負の数を指定すると、デクリメンタとしても使える。
 */
template <typename T>
class Incrementer
{
    size_t count_;
    int subtract_;
    const T begin_, increment_, min_value_, max_value_;

public:
    /** @brief 初期値としてbeginを持つ増分器を生成する.
     * @param begin 増分器の初期値
     * @param increment 一回の増加量
     * @param min_value 増加範囲の下界
     * @param max_value 増加範囲の上界
     */
    Incrementer(T begin, T increment, T min_value = NumericLimitsLowest<T>(), T max_value = std::numeric_limits<T>::max())
        : count_(0), subtract_(0),
          begin_(begin), increment_(increment), min_value_(min_value), max_value_(max_value)
    {}

    /** @brief incrementだけインクリメントする.
     * @return インクリメントされたオブジェクト
     */
    Incrementer& operator ++()
    {
        count_++;
        if (CalcCurrentAngle() >= max_value_)
        {
            subtract_++;
        }
        else if (CalcCurrentAngle() < min_value_)
        {
            subtract_--;
        }
        return *this;
    }

    /** @brief incrementだけインクリメントする.
     * @return インクリメントする前のオブジェクト
     */
    Incrementer operator ++(int)
    {
        Incrementer temp = *this;
        this->operator ++();
        return temp;
    }

    /** @brief 現在の数値を返す.
     * @return 現在の数値
     */
    T operator *() const
    {
        return CalcCurrentAngle();
    }

    /** @brief 初期状態に戻す.
     */
    void Reset()
    {
        this->count_ = 0;
        this->subtract_ = 0;
    }

private:
    T CalcCurrentAngle() const
    {
        return begin_ + increment_ * count_ - (max_value_ - min_value_) * subtract_;
    }
};

}

#endif // HARKTOOL4_INCREMENTER_HPP

